#  Copyright (c) 2021 Food-X Technologies
#
#  This file is part of foodx_devops_tools.
#
#  You should have received a copy of the MIT License along with
#  foodx_devops_tools. If not, see <https://opensource.org/licenses/MIT>.

import asyncio
import logging
import pathlib
import sys
import typing

import click

from foodx_devops_tools._declarations import (
    DEFAULT_CONSOLE_LOGGING_ENABLED,
    DEFAULT_FILE_LOGGING_DISABLED,
    DEFAULT_LOG_LEVEL,
    VALID_LOG_LEVELS,
)
from foodx_devops_tools._logging import LoggingState
from foodx_devops_tools._to import StructuredTo, StructuredToParameter
from foodx_devops_tools._version import acquire_version
from foodx_devops_tools.pipeline_config import (
    DeploymentContext,
    PipelineConfiguration,
    PipelineConfigurationPaths,
    ReleaseView,
)
from foodx_devops_tools.pipeline_config.exceptions import (
    ConfigurationPathsError,
)
from foodx_devops_tools.release_flow import (
    identify_release_id,
    identify_release_state,
)
from foodx_devops_tools.utilities import acquire_token, get_sha

from ._deployment import (
    DeploymentState,
    FlattenedDeployment,
    assess_results,
    do_deploy,
)
from ._exceptions import DeploymentTerminatedError
from ._state import ExitState, PipelineCliOptions

log = logging.getLogger(__name__)

DEFAULT_LOG_FILE = pathlib.Path("deploy_me.log")


class DeploymentConfigurationError(Exception):
    """Problem acquiring deployment configuration."""


async def _gather_main(
    configuration: PipelineConfiguration,
    deployment_iterations: typing.List[FlattenedDeployment],
    pipeline_parameters: PipelineCliOptions,
) -> None:
    """Deploy each deployment iteration asynchronously."""
    results = await asyncio.gather(
        *[
            do_deploy(configuration, x, pipeline_parameters)
            for x in deployment_iterations
        ],
        return_exceptions=False,
    )

    filtered_results = [x for x in results if isinstance(x, DeploymentState)]
    if len(filtered_results) != len(results):
        log.error("Some deployments may have had unexpected failures.")

    condensed_result = await assess_results(filtered_results)
    _report_results(condensed_result.code, len(deployment_iterations))


def _report_results(
    result_code: DeploymentState.ResultType, number_iterations: int
) -> None:
    if (result_code == DeploymentState.ResultType.success) and (
        number_iterations != 0
    ):
        click.echo(click.style("success: Deployment succeeded.", fg="green"))
    elif (result_code == DeploymentState.ResultType.success) and (
        number_iterations == 0
    ):
        click.echo(click.style("skipped: Deployment skipped.", fg="yellow"))
        # exit "clean" because a skip is not a failure per-se.
    else:
        click.echo(
            click.style(
                "FAILED: Deployment failed. Check log for details.",
                fg="red",
            )
        )
        sys.exit(ExitState.DEPLOYMENT_FAILED.value)


@click.command()
@click.version_option(acquire_version())
@click.argument(
    "client_config",
    type=click.Path(dir_okay=True, file_okay=False, path_type=pathlib.Path),
)
@click.argument(
    "system_config",
    type=click.Path(dir_okay=True, file_okay=False, path_type=pathlib.Path),
)
@click.argument(
    "password_file",
    type=click.File(mode="r"),
)
@click.option(
    "--git-ref",
    default=None,
    help="""Git reference to drive release management workflow.

eg.

    ``$(Build.SourceBranch)`` for Azure DevOps Pipelines.
    ``${GITHUB_REF}`` for Github Actions.
    ``${CI_COMMIT_REF_NAME}`` for Gitlab-CI.
""",
    type=str,
)
@click.option(
    "--log-enable-console",
    "enable_console_log",
    default=DEFAULT_CONSOLE_LOGGING_ENABLED,
    help="Log to console.",
    is_flag=True,
)
@click.option(
    "--log-disable-file",
    "disable_file_log",
    default=DEFAULT_FILE_LOGGING_DISABLED,
    help="Disable file logging.",
    is_flag=True,
)
@click.option(
    "--log-level",
    "log_level",
    default=DEFAULT_LOG_LEVEL,
    help="Select logging level to apply to all enabled log sinks.",
    show_default=True,
    type=click.Choice(VALID_LOG_LEVELS, case_sensitive=False),
)
@click.option(
    "--monitor-sleep",
    default=30,
    help="Sleep time in seconds between frame dependency status polls.",
    show_default=True,
    type=int,
)
@click.option(
    "--pipeline-id",
    default="000+local",
    help="""The pipeline ID. Should be considered mandatory for a CI/CD
 pipeline (optional for developers running locally).

eg.

    ``$(Build.BuildNumber)`` for Azure DevOps Pipelines.
    ``${GITHUB_RUN_NUMBER}`` for Github Actions.
    ``${CI_PIPELINE_ID}`` for Gitlab-CI.""",
    type=str,
)
@click.option(
    "--to",
    default=None,
    help="""Specify a structured name to deploy a specific system component.

<frame>.<application>.<step>

[default: deploy everything]
""",
    type=StructuredToParameter(),
)
@click.option(
    "--validation",
    default=False,
    help="Force deployments to be a validation deployment, regardless of any "
    "specified deployment mode in configuration.",
    is_flag=True,
)
@click.option(
    "--wait-timeout",
    default=15,
    help="Maximum wait time in minutes for frame, application or step "
    "completion. Failure results if the timeout is exceeded.",
    show_default=True,
    type=int,
)
def deploy_me(
    client_config: pathlib.Path,
    system_config: pathlib.Path,
    password_file: typing.IO,
    disable_file_log: bool,
    enable_console_log: bool,
    log_level: str,
    monitor_sleep: int,
    git_ref: typing.Optional[str],
    pipeline_id: str,
    to: StructuredTo,
    validation: bool,
    wait_timeout: int,
) -> None:
    """
    Deploy system resources.

    CLIENT_CONFIG  The client specific configuration directory.
    SYSTEM_CONFIG  The directory containing all non-client related pipeline
                   and deployment configuration.
    PASSWORD_FILE:  The path to a file where the service principal decryption
                    password is stored, or "-" for stdin.
    """
    try:
        # currently no need to change logging configuration at run time,
        # so no need to preserve the object.
        LoggingState(
            disable_file_logging=disable_file_log,
            enable_console_logging=enable_console_log,
            log_level_text=log_level,
            default_log_file=DEFAULT_LOG_FILE,
        )

        pipeline_parameters = PipelineCliOptions(
            enable_validation=validation,
            monitor_sleep_seconds=monitor_sleep,
            wait_timeout_seconds=(60 * wait_timeout),
        )
        configuration_paths = PipelineConfigurationPaths.from_paths(
            client_config, system_config
        )
        decrypt_token = acquire_token(password_file)
        this_configuration = PipelineConfiguration.from_files(
            configuration_paths, decrypt_token
        )

        if git_ref:
            release_id = identify_release_id(git_ref)
            release_state = identify_release_state(git_ref)
        else:
            raise DeploymentConfigurationError(
                "--git-ref is mandatory (for now)"
            )

        commit_sha = get_sha()
        base_context = DeploymentContext(
            commit_sha=commit_sha,
            pipeline_id=pipeline_id,
            release_id=release_id,
            release_state=release_state.name,
        )
        log.info("top-level deployment context, {0}".format(str(base_context)))

        pipeline_state = ReleaseView(this_configuration, base_context)
        deployment_iterations = pipeline_state.flatten(to)

        log.info(
            "number deployment iteration, {0}".format(
                len(deployment_iterations)
            )
        )
        log.debug(str(deployment_iterations))

        asyncio.run(
            _gather_main(
                this_configuration, deployment_iterations, pipeline_parameters
            )
        )
    except (ConfigurationPathsError, DeploymentConfigurationError) as e:
        message = str(e)
        log.error(message)
        click.echo(message, err=True)
        sys.exit(ExitState.BAD_DEPLOYMENT_CONFIGURATION.value)
    except asyncio.CancelledError:
        log.error("Async cancellation exception")
        click.echo("Exiting due to async cancellation", err=True)
        sys.exit(ExitState.DEPLOYMENT_CANCELLED.value)
    except DeploymentTerminatedError as e:
        log.error("Deployment cancelled exception, {0}".format(str(e)))
        click.echo(
            click.style("Exiting due to deployment cancellation", fg="red"),
            err=True,
        )
        sys.exit(ExitState.DEPLOYMENT_CANCELLED.value)
    except asyncio.TimeoutError:
        click.echo(click.style("Exiting due to timeout", fg="red"), err=True)
        sys.exit(ExitState.DEPLOYMENT_TIMEOUT.value)
    except Exception as e:
        log.exception(str(e))
        click.echo(
            "Deployment failed with unexpected error (see log for "
            "details), {0}".format(str(e)),
            err=True,
        )
        sys.exit(ExitState.UNKNOWN_ERROR.value)



#  Copyright (c) 2021 Food-X Technologies
#
#  This file is part of foodx_devops_tools.
#
#  You should have received a copy of the MIT License along with
#  foodx_devops_tools. If not, see <https://opensource.org/licenses/MIT>.
#
#  This file is part of foodx_devops_tools.
#
#  You should have received a copy of the MIT License along with
#  foodx_devops_tools. If not, see <https://opensource.org/licenses/MIT>.

import pytest

from foodx_devops_tools.deploy_me._deployment import (
    DeploymentState,
    _construct_fqdn,
    _construct_resource_group_name,
    assess_results,
)


class TestConstructFqdn:
    def test_structured_subscription(self):
        result = _construct_fqdn(
            "api", "some.where", "client", "system_client_stub"
        )

        assert result == "api.stub.client.some.where"

    def test_unstructured_subscription(self):
        result = _construct_fqdn(
            "api", "some.where", "client", "system_client_stub"
        )

        assert result == "api.stub.client.some.where"


class TestConstructResourceGroupName:
    def test_clean(self):
        result = _construct_resource_group_name("app", "frame", "client")

        assert result == "app-frame-client"


class TestAssessResults:
    @pytest.mark.asyncio
    async def test_success(self):
        mock_results = [
            DeploymentState(code=DeploymentState.ResultType.success),
            DeploymentState(code=DeploymentState.ResultType.success),
        ]

        result = await assess_results(mock_results)

        assert result.code == DeploymentState.ResultType.success

    @pytest.mark.asyncio
    async def test_fail(self):
        mock_results = [
            DeploymentState(code=DeploymentState.ResultType.success),
            DeploymentState(code=DeploymentState.ResultType.failed),
        ]

        result = await assess_results(mock_results)

        assert result.code == DeploymentState.ResultType.failed



#  Copyright (c) 2021 Food-X Technologies
#
#  This file is part of foodx_devops_tools.
#
#  You should have received a copy of the MIT License along with
#  foodx_devops_tools. If not, see <https://opensource.org/licenses/MIT>.

import copy
import pathlib
import typing
import unittest.mock

import pydantic
import pytest
from asynctest import CoroutineMock

from foodx_devops_tools.pipeline_config import PipelineConfiguration
from foodx_devops_tools.pipeline_config.views import (
    DeploymentContext,
    ReleaseView,
)
from tests.ci.support.pipeline_config import MOCK_RESULTS, MOCK_TO


@pytest.fixture
def apply_pipeline_config_test(mocker):
    def _apply(
        mock_content: str,
        method_under_test: typing.Callable[[pathlib.Path], pydantic.BaseModel],
    ):
        mock_path = mocker.create_autospec(pathlib.Path)("some/path")
        mock_path.open = mocker.mock_open(read_data=mock_content)

        result = method_under_test(mock_path)

        return result

    return _apply


@pytest.fixture
def mock_pipeline_config():
    def _apply(mock_data=copy.deepcopy(MOCK_RESULTS)) -> PipelineConfiguration:
        return PipelineConfiguration.parse_obj(mock_data)

    return _apply


@pytest.fixture()
def mock_flattened_deployment(mock_pipeline_config):
    base_context = DeploymentContext(
        commit_sha="abc123",
        pipeline_id="123456",
        release_id="3.1.4+local",
        release_state="r1",
    )
    pipeline_state = ReleaseView(mock_pipeline_config(), base_context)
    mock_flattened = pipeline_state.flatten(MOCK_TO)

    return copy.deepcopy(mock_flattened)


@pytest.fixture()
def mock_async_method(mocker):
    def _apply(
        path_to_mock: str,
        return_value: typing.Optional[typing.Any] = None,
        side_effect: typing.Optional[typing.Any] = None,
    ):
        async_mock = unittest.mock.AsyncMock(
            return_value=return_value, side_effect=side_effect
        )
        this_mock = mocker.patch(path_to_mock, side_effect=async_mock)

        return this_mock

    return _apply


@pytest.fixture()
def mock_context(mocker):
    def _apply(path_to_mock: str):
        this_mock = mocker.patch(path_to_mock)
        this_mock.return_value.__aenter__.return_value.write = CoroutineMock()

        return this_mock

    return _apply



#  Copyright (c) 2021 Food-X Technologies
#
#  This file is part of foodx_devops_tools.
#
#  You should have received a copy of the MIT License along with
#  foodx_devops_tools. If not, see <https://opensource.org/licenses/MIT>.

import os

import pytest

from foodx_devops_tools.puff.arm import load_yaml


class TestLoadYaml:
    @pytest.mark.asyncio
    async def test_clean(self, tmp_path_factory):
        this_directory = tmp_path_factory.mktemp("tmp")
        this_path = this_directory / "some.file"
        with this_path.open(mode="w") as f:
            f.write(
                """---
some:
- d1
- d2
"""
            )
        result = await load_yaml(this_path)

        assert result == {"some": ["d1", "d2"]}

    @pytest.mark.asyncio
    async def test_empty_yaml_warns(self, caplog, tmp_path_factory):
        this_directory = tmp_path_factory.mktemp("tmp")
        this_path = this_directory / "some.file"
        with this_path.open(mode="w") as f:
            f.write(
                """---
"""
            )

        result = await load_yaml(this_path)

        assert any(["Empty YAML file" in x for x in caplog.messages])

    @pytest.mark.asyncio
    async def test_missing_file_raises(self, caplog, tmp_path_factory):
        this_directory = tmp_path_factory.mktemp("tmp")
        this_path = this_directory / "some.file"

        with pytest.raises(
            FileNotFoundError, match=r"No such file or " r"directory"
        ):
            await load_yaml(this_path)

        assert any(["File not found" in x for x in caplog.messages])

    @pytest.mark.asyncio
    async def test_directory_raises(self, caplog, tmp_path_factory):
        """Path pointing to a directory raises an exception."""
        this_directory = tmp_path_factory.mktemp("tmp")
        this_path = this_directory / "some.dir"
        os.makedirs(this_path, exist_ok=True)

        with pytest.raises(IsADirectoryError, match=r"Is a directory"):
            await load_yaml(this_path)

        assert any(["Path is a directory" in x for x in caplog.messages])

    @pytest.mark.asyncio
    async def test_unexpected_logs(self, caplog, mocker, tmp_path_factory):
        """Unexpected error logs a report."""
        this_directory = tmp_path_factory.mktemp("tmp")
        mocker.patch(
            "foodx_devops_tools.puff.arm.ruamel.yaml.YAML",
            side_effect=RuntimeError("some error"),
        )
        this_path = this_directory / "some.dir"
        os.makedirs(this_path, exist_ok=True)

        with pytest.raises(RuntimeError, match=r"^some error"):
            await load_yaml(this_path)

        assert any(["Unexpected error with path" in x for x in caplog.messages])



#  Copyright (c) 2021 Food-X Technologies
#
#  This file is part of foodx_devops_tools.
#
#  You should have received a copy of the MIT License along with
#  foodx_devops_tools. If not, see <https://opensource.org/licenses/MIT>.

"""Core implementation of ``puff`` utility."""

import asyncio
import dataclasses
import enum
import logging
import pathlib
import typing

import click

from ._ascii_art import JELLY, PUFFIN
from ._exceptions import PuffError
from .arm import do_arm_template_parameter_action
from .puffignore import IgnorePatterns, load_puffignore

log = logging.getLogger(__name__)

GENERATING_MESSAGE = "generating files..."
DELETING_MESSAGE = "deleting generated files..."


@enum.unique
class PuffActions(enum.Enum):
    """Puff utility action types."""

    create = enum.auto()
    delete = enum.auto()


@dataclasses.dataclass
class ActionConfiguration:
    """Configuration defining puff actions."""

    message: str
    art: str


ACTION_MESSAGES = {
    PuffActions.create: ActionConfiguration(
        message=GENERATING_MESSAGE, art=JELLY
    ),
    PuffActions.delete: ActionConfiguration(
        message=DELETING_MESSAGE, art=PUFFIN
    ),
}
DEFAULT_PUFFIGNORE_PATH = pathlib.Path("./.puffignore")

YamlFilenames = typing.Set[pathlib.Path]


def _acquire_yaml_filenames(
    path: pathlib.Path, ignore_patterns: IgnorePatterns
) -> YamlFilenames:
    """
    Collect paths to YAML files under root path.

    Args:
        path: Root path to search recursively.
        ignore_patterns: Unix style globs to exclude from search results.

    Returns:
        Discovered YAML file names excluding those matching ignore patterns.
    """
    yaml_files = set(path.glob("**/*.yml")).union(set(path.glob("**/*.yaml")))
    for this_pattern in ignore_patterns:
        yaml_files = yaml_files.difference(path.glob(this_pattern))

    return yaml_files


async def run_puff(
    path: pathlib.Path, is_delete_files: bool, is_pretty: bool
) -> None:
    """
    Search filesystem for YAML files and create or delete ARM template files.

    Args:
        path: Root path to search recursively for YAML files.
        is_delete_files: Enable/disable delete instead of create action.
        is_pretty: Create nicely formatted JSON for humans.
    """
    if is_delete_files:
        this_action = PuffActions.delete
    else:
        this_action = PuffActions.create

    click.echo(ACTION_MESSAGES[this_action].art)
    click.echo(ACTION_MESSAGES[this_action].message)

    ignore_patterns = await load_puffignore(DEFAULT_PUFFIGNORE_PATH)
    if path.is_dir():
        yaml_filenames = _acquire_yaml_filenames(path, ignore_patterns)
    elif path.is_file():
        yaml_filenames = {path}
    else:
        raise PuffError("Path must be file or directory, {0}".format(path))

    log.debug(str(yaml_filenames))

    if not yaml_filenames:
        log.warning(
            "No puff parameter files found in directory, {0}".format(path)
        )

    await asyncio.gather(
        *[
            do_arm_template_parameter_action(x, is_delete_files, is_pretty)
            for x in yaml_filenames
        ]
    )



#  Copyright (c) 2021 Food-X Technologies
#
#  This file is part of foodx_devops_tools.
#
#  You should have received a copy of the MIT License along with
#  foodx_devops_tools. If not, see <https://opensource.org/licenses/MIT>.
#
#  This file is part of foodx_devops_tools.
#
#  You should have received a copy of the MIT License along with
#  foodx_devops_tools. If not, see <https://opensource.org/licenses/MIT>.

"""Manage the structured ``--to`` specifier."""

import dataclasses
import logging
import re
import typing

import click

log = logging.getLogger(__name__)

SPECIFIER_REGEX = (
    r"^(?P<frame>[a-z0-9_\-]+)"
    r"(\."
    r"(?P<application>[a-z0-9_\-]+)"
    r"(\."
    r"(?P<step>[a-z0-9_\-]+)"
    r")?"
    r")?$"
)


class StructuredToError(Exception):
    """Problem occurred parsing a "--to" specifier."""


U = typing.TypeVar("U", bound="StructuredTo")


@dataclasses.dataclass()
class StructuredTo:
    """Structured deployment specifier."""

    frame: typing.Optional[str] = None
    application: typing.Optional[str] = None
    step: typing.Optional[str] = None

    def __str__(self: U) -> str:
        """Convert structure back to "specifier" form for use in messages."""
        if (not self.frame) and (not self.application) and (not self.step):
            result = "<all>"
        elif (not self.application) and (not self.step):
            result = "{0}".format(self.frame)
        elif not self.step:
            result = "{0}.{1}".format(self.frame, self.application)
        else:
            result = "{0}.{1}.{2}".format(
                self.frame, self.application, self.step
            )
        return result

    @classmethod
    def from_specifier(
        cls: typing.Type[U], to_specified: typing.Optional[str]
    ) -> U:
        """
        Construct a ``StructuredTo`` object from a specifier string.

        Args:
            to_specified: Structured specifier string to be parsed.

        Returns:
            Constructed ``StructuredTo`` object.
        Raises:
            StructuredToError: If the string specified is non conformant.
        """
        this_object = cls()

        if to_specified:
            this_match = re.match(SPECIFIER_REGEX, to_specified)
            if this_match:
                this_frame = this_match.group("frame")
                if this_frame:
                    this_object.frame = this_frame
                    this_application = this_match.group("application")
                    if this_application:
                        this_object.application = this_application
                        this_step = this_match.group("step")
                        if this_step:
                            this_object.step = this_step
            else:
                message = "Bad structured deployment specifier, {0}".format(
                    to_specified
                )
                raise StructuredToError(message)

        return this_object


T = typing.TypeVar("T", bound="StructuredToParameter")


class StructuredToParameter(click.ParamType):
    """Custom click parameter for the structured to option."""

    # https://click.palletsprojects.com/en/8.0.x/parameters/#implementing-custom-types
    def convert(
        self: T,
        value: typing.Optional[typing.Union[str, StructuredTo]],
        param: typing.Optional[click.Parameter],
        context: typing.Optional[click.Context],
    ) -> StructuredTo:
        """Convert command line option value to a structured to type."""
        if isinstance(value, StructuredTo):
            return value

        try:
            if isinstance(value, str) or (value is None):
                this_object = StructuredTo.from_specifier(value)
            else:
                raise StructuredToError()

            return this_object
        except StructuredToError:
            self.fail(
                f"{value!r} is not a valid deployment specifier", param, context
            )



#  Copyright (c) 2021 Food-X Technologies
#
#  This file is part of foodx_devops_tools.
#
#  You should have received a copy of the MIT License along with
#  foodx_devops_tools. If not, see <https://opensource.org/licenses/MIT>.

"""Systems deployment configuration I/O."""

import pathlib
import typing

import pydantic

from ._exceptions import SystemsDefinitionError
from ._loader import load_configuration

ENTITY_NAME = "systems"

T = typing.TypeVar("T", bound="SystemsDefinition")

ValueType = typing.List[str]


class SystemsDefinition(pydantic.BaseModel):
    """Definition of deployable systems."""

    systems: ValueType

    @pydantic.validator(ENTITY_NAME)
    def check_systems(cls: pydantic.BaseModel, value: ValueType) -> ValueType:
        """Validate ``systems`` field."""
        if not value:
            raise ValueError("Empty system names prohibited")
        if len(set(value)) != len(value):
            raise ValueError("Duplicate system names prohibited")

        return value


def load_systems(file_path: pathlib.Path) -> SystemsDefinition:
    """
    Load system definitions from file.

    Args:
        file_path: Path to system definitions file.

    Returns:
        System definitions.
    Raises:
        SystemsDefinitionError: If an error occurs loading the file.
    """
    result = load_configuration(
        file_path, SystemsDefinition, SystemsDefinitionError, ENTITY_NAME
    )

    return typing.cast(SystemsDefinition, result)



#  Copyright (c) 2021 Food-X Technologies
#
#  This file is part of foodx_devops_tools.
#
#  You should have received a copy of the MIT License along with
#  foodx_devops_tools. If not, see <https://opensource.org/licenses/MIT>.

"""Manage .puffignore data and methods."""

import copy
import logging
import pathlib
import typing

import aiofiles  # type: ignore

log = logging.getLogger(__name__)

BASE_IGNORE = ["**/node_modules/**/*"]

IgnorePatterns = typing.List[str]


async def load_puffignore(puffignore_path: pathlib.Path) -> IgnorePatterns:
    """
    Load .puffignore data from current working directory.

    A non-existent or empty .puffignore file is silently ignored.

    Returns:
        Loaded ignore patterns or empty list.
    """
    try:
        async with aiofiles.open(str(puffignore_path), mode="r") as f:
            lines = await f.readlines()
    except FileNotFoundError:
        log.info(
            "No .puffignore file found, {0}".format(puffignore_path.absolute())
        )
        lines = list()

    filtered_lines = copy.deepcopy(BASE_IGNORE)
    for this_line in lines:
        stripped_line = this_line.strip()
        if stripped_line and not stripped_line.startswith("#"):
            if stripped_line.endswith("/"):
                # append wildcard
                filtered_lines.append(stripped_line + "*")
            else:
                filtered_lines.append(stripped_line)

    return filtered_lines



#  Copyright (c) 2021 Food-X Technologies
#
#  This file is part of foodx_devops_tools.
#
#  You should have received a copy of the MIT License along with
#  foodx_devops_tools. If not, see <https://opensource.org/licenses/MIT>.

import contextlib
import copy
import json
import pathlib
import tempfile

from foodx_devops_tools.puff_utility import _main
from tests.ci.support.click_runner import (  # noqa: F401
    click_runner,
    isolated_filesystem,
)

SAMPLE_DATA = """services:
  s1:
    p1: s1p1
    environments:
      e1:
        p1: s1e1p1
        regions:
          - r1:
              p1: s1e1r1p1
              p2:
                reference:
                  keyVault:
                    id: idv
                  secretName: sn
"""


@contextlib.contextmanager
def mock_file(
    file_name: str,
    content: str,
):
    with tempfile.TemporaryDirectory() as dir:
        expected_dir = pathlib.Path(dir)
        file_path = expected_dir / file_name
        with file_path.open(mode="w") as f:
            f.write(content)

        yield file_path


def test_generate_clean(click_runner):
    """Empty ignore patterns run acquires all YAML files."""
    expected_data = {
        "$schema": "https://schema.management.azure.com/schemas/2019-04-01/deploymentParameters.json#",
        "contentVersion": "1.0.0.0",
        "parameters": {
            "service": {"value": "s1"},
            "environment": {"value": "e1"},
            "region": {"value": "r1"},
            "p1": {"value": "s1e1r1p1"},
            "p2": {
                "reference": {
                    "keyVault": {
                        "id": "idv",
                    },
                    "secretName": "sn",
                },
            },
        },
    }

    with mock_file("some.yml", copy.deepcopy(SAMPLE_DATA)) as file_path:
        click_runner.invoke(_main, [str(file_path.parent)])

        expected_file = file_path.parent / "some.s1.e1.r1.json"
        with expected_file.open("r") as f:
            data = json.load(f)

        assert data == expected_data



#  Copyright (c) 2021 Food-X Technologies
#
#  This file is part of foodx_devops_tools.
#
#  You should have received a copy of the MIT License along with
#  foodx_devops_tools. If not, see <https://opensource.org/licenses/MIT>.

"""Generalized structured name for deep referencing objects in data model."""

import pathlib
import typing

T = typing.TypeVar("T", bound="StructuredName")


class StructuredName(typing.List[str]):
    """Structured name of frames, applications and application steps."""

    def __hash__(self: T) -> int:  # type: ignore[override]
        """Hash ``StructuredName`` object for use as dictionary keys."""
        return hash(str(self))

    def __str__(self: T) -> str:
        """Convert ``StructuredName`` object to a string representation."""
        return ".".join(self)


StructuredPathCollection = typing.Dict[StructuredName, pathlib.Path]



#  Copyright (c) 2021 Food-X Technologies
#
#  This file is part of foodx_devops_tools.
#
#  You should have received a copy of the MIT License along with
#  foodx_devops_tools. If not, see <https://opensource.org/licenses/MIT>.

import copy

import pytest

from foodx_devops_tools.pipeline_config.exceptions import PipelineViewError
from foodx_devops_tools.pipeline_config.views import (
    DeploymentTuple,
    DeploymentView,
    IterationContext,
    ReleaseView,
    SubscriptionView,
)
from tests.ci.support.pipeline_config import MOCK_CONTEXT, MOCK_TO


class TestIterationContext:
    def test_clean(self):
        under_test = IterationContext()
        under_test.append("one")
        under_test.append("two")

        assert str(under_test) == "one.two"

    def test_empty(self):
        under_test = IterationContext()

        assert str(under_test) == ""


class TestFlattenedDeployment:
    def test_copy_add_frame(self, mock_flattened_deployment):
        expected_frame = "f1"
        under_test = mock_flattened_deployment[0]

        assert expected_frame not in under_test.data.iteration_context

        result = under_test.copy_add_frame(expected_frame)

        assert id(result) != id(under_test)
        assert expected_frame in result.data.iteration_context
        assert result.context.frame_name == expected_frame

    def test_copy_add_application(self, mock_flattened_deployment):
        expected_name = "a1"
        under_test = mock_flattened_deployment[0]

        assert expected_name not in under_test.data.iteration_context

        result = under_test.copy_add_application(expected_name)

        assert id(result) != id(under_test)
        assert expected_name in result.data.iteration_context
        assert result.context.application_name == expected_name


class TestSubscriptionView:
    def test_clean(self, mock_pipeline_config):
        release_view = ReleaseView(mock_pipeline_config(), MOCK_CONTEXT)
        deployment_view = DeploymentView(
            release_view,
            DeploymentTuple(client="c1", release_state="r1", system="sys1"),
        )
        under_test = SubscriptionView(
            deployment_view,
            "sub1",
        )

    def test_deploy_data_clean(self, mock_pipeline_config):
        release_view = ReleaseView(mock_pipeline_config(), MOCK_CONTEXT)
        deployment_view = DeploymentView(
            release_view,
            DeploymentTuple(client="c1", release_state="r1", system="sys1"),
        )
        under_test = SubscriptionView(
            deployment_view,
            "sub1",
        )

        result = under_test.deploy_data

        assert [x.azure_credentials.userid for x in result] == [
            "12345",
            "12345",
        ]
        assert {x.location_primary for x in result} == {"l1", "l2"}

    def test_bad_subscription_raises(self, mock_pipeline_config):
        release_view = ReleaseView(mock_pipeline_config(), MOCK_CONTEXT)
        deployment_view = DeploymentView(
            release_view,
            DeploymentTuple(client="c1", release_state="r1", system="sys1"),
        )
        with pytest.raises(PipelineViewError, match=r"^Bad subscription"):
            SubscriptionView(
                deployment_view,
                "bad_sub",
            )


class TestDeploymentView:
    def test_clean(self, mock_pipeline_config):
        expected_state = DeploymentTuple(
            client="c1", release_state="r2", system="sys1"
        )
        this_context = copy.deepcopy(MOCK_CONTEXT)
        this_context.release_state = "r2"
        release_view = ReleaseView(mock_pipeline_config(), this_context)
        under_test = DeploymentView(release_view, expected_state)

        assert under_test.release_view == release_view
        assert under_test.deployment_tuple == expected_state

    def test_bad_deployment_state_raises(self, mock_pipeline_config):
        this_context = copy.deepcopy(MOCK_CONTEXT)
        this_context.release_state = "r2"
        release_view = ReleaseView(mock_pipeline_config(), this_context)
        with pytest.raises(PipelineViewError, match=r"^Bad client"):
            DeploymentView(
                release_view,
                DeploymentTuple(
                    client="bad_client", release_state="r2", system="sys2"
                ),
            )
        with pytest.raises(PipelineViewError, match=r"^Bad release state"):
            DeploymentView(
                release_view,
                DeploymentTuple(
                    client="c1", release_state="bad_release", system="sys2"
                ),
            )
        with pytest.raises(PipelineViewError, match=r"^Bad system"):
            DeploymentView(
                release_view,
                DeploymentTuple(
                    client="c1", release_state="r2", system="bad_system"
                ),
            )

    def test_subscriptions_clean(self, mock_pipeline_config):
        expected_state = DeploymentTuple(
            client="c1", release_state="r1", system="sys1"
        )
        this_context = copy.deepcopy(MOCK_CONTEXT)
        this_context.release_state = "r2"
        release_view = ReleaseView(mock_pipeline_config(), this_context)
        under_test = DeploymentView(release_view, expected_state)

        assert under_test.release_view == release_view
        assert under_test.deployment_tuple == expected_state

        assert under_test.subscriptions

    def test_subscriptions_empty(self, mock_pipeline_config):
        this_context = copy.deepcopy(MOCK_CONTEXT)
        this_context.release_state = "r2"
        release_view = ReleaseView(mock_pipeline_config(), this_context)
        under_test = DeploymentView(
            release_view,
            DeploymentTuple(client="c2", release_state="r2", system="sys2"),
        )

        assert not under_test.subscriptions


class TestReleaseView:
    def test_clean(self, mock_pipeline_config):
        this_context = copy.deepcopy(MOCK_CONTEXT)
        this_context.release_state = "r2"
        under_test = ReleaseView(mock_pipeline_config(), this_context)

        assert under_test.configuration == mock_pipeline_config()
        assert under_test.deployment_context.release_state == "r2"

    def test_bad_release_state_raises(self, mock_pipeline_config):
        this_context = copy.deepcopy(MOCK_CONTEXT)
        this_context.release_state = "r3"
        with pytest.raises(PipelineViewError, match=r"^Bad release state"):
            ReleaseView(mock_pipeline_config(), this_context)

    def test_deployments_clean(self, mock_pipeline_config):
        under_test = ReleaseView(mock_pipeline_config(), MOCK_CONTEXT)

        result = under_test.deployments

        result_tuples = {str(x.deployment_tuple) for x in result}
        assert result_tuples == {
            "sys1-c1-r1",
        }

    def test_deployments_empty(self, mock_pipeline_config):
        this_context = copy.deepcopy(MOCK_CONTEXT)
        this_context.release_state = "r2"
        under_test = ReleaseView(mock_pipeline_config(), this_context)

        result = under_test.deployments

        result_tuples = {str(x.deployment_tuple) for x in result}
        assert not result_tuples

    def test_flatten_clean(self, mock_pipeline_config):
        under_test = ReleaseView(mock_pipeline_config(), MOCK_CONTEXT)

        result = under_test.flatten(MOCK_TO)

        assert len(result) == 2
        assert {x.context.client for x in result} == {
            "c1",
        }
        assert {x.context.system for x in result} == {
            "sys1",
        }
        assert {x.data.location_primary for x in result} == {
            "l1",
            "l2",
        }
        assert [x.data.to for x in result] == [
            MOCK_TO,
            MOCK_TO,
        ]



#  Copyright (c) 2021 Food-X Technologies
#
#  This file is part of foodx_devops_tools.
#
#  You should have received a copy of the MIT License along with
#  foodx_devops_tools. If not, see <https://opensource.org/licenses/MIT>.

"""Context sensitive views of pipeline configuration data."""

import copy
import dataclasses
import logging
import pathlib
import typing

from foodx_devops_tools._to import StructuredTo
from foodx_devops_tools.azure.cloud import AzureCredentials

from ..deployment import DeploymentTuple
from ._exceptions import PipelineViewError
from .pipeline import PipelineConfiguration, PuffMap

log = logging.getLogger(__name__)

Z = typing.TypeVar("Z", bound="IterationContext")


class IterationContext(typing.List[str]):
    """Record the iteration path."""

    def __str__(self: Z) -> str:
        """Convert the iteration path to a structured string."""
        return ".".join(self)


Y = typing.TypeVar("Y", bound="DeploymentContext")


class DeploymentContext:
    """Deployment context data expected to be applied to resource tags."""

    commit_sha: str
    pipeline_id: str
    release_id: str
    release_state: str

    # these data must be provided at deployment time because they are more
    # deeply context sensitive.
    __application_name: typing.Optional[str] = None
    __azure_subscription_name: typing.Optional[str] = None
    __azure_tenant_name: typing.Optional[str] = None
    __client: typing.Optional[str] = None
    __frame_name: typing.Optional[str] = None
    __system: typing.Optional[str] = None

    def __init__(
        self: Y,
        commit_sha: str,
        pipeline_id: str,
        release_id: str,
        release_state: str,
    ) -> None:
        """Construct ``DeploymentContext`` object."""
        self.commit_sha = commit_sha
        self.pipeline_id = pipeline_id
        self.release_id = release_id
        self.release_state = release_state

    @property
    def application_name(self: Y) -> str:
        """Get application name property."""
        if self.__application_name:
            return self.__application_name
        else:
            raise PipelineViewError(
                "application name not specified in deployment data"
            )

    @application_name.setter
    def application_name(self: Y, v: str) -> None:
        """Set application name property."""
        self.__application_name = v

    @property
    def azure_subscription_name(self: Y) -> str:
        """Get azure subscription name property."""
        if self.__azure_subscription_name:
            return self.__azure_subscription_name
        else:
            raise PipelineViewError(
                "azure subscription name not specified in deployment data"
            )

    @azure_subscription_name.setter
    def azure_subscription_name(self: Y, v: str) -> None:
        """Set azure subscription name property."""
        self.__azure_subscription_name = v

    @property
    def azure_tenant_name(self: Y) -> str:
        """Get azure tenant name property."""
        if self.__azure_tenant_name:
            return self.__azure_tenant_name
        else:
            raise PipelineViewError(
                "azure tenant name not specified in deployment data"
            )

    @azure_tenant_name.setter
    def azure_tenant_name(self: Y, v: str) -> None:
        """Set azure tenant name property."""
        self.__azure_tenant_name = v

    @property
    def client(self: Y) -> str:
        """Get client name property."""
        if self.__client:
            return self.__client
        else:
            raise PipelineViewError("client not specified in deployment data")

    @client.setter
    def client(self: Y, v: str) -> None:
        """Set client name property."""
        self.__client = v

    @property
    def frame_name(self: Y) -> str:
        """Get frame name property."""
        if self.__frame_name:
            return self.__frame_name
        else:
            raise PipelineViewError(
                "frame name not specified in deployment data"
            )

    @frame_name.setter
    def frame_name(self: Y, v: str) -> None:
        """Set frame name property."""
        self.__frame_name = v

    @property
    def system(self: Y) -> str:
        """Get system name property."""
        if self.__system:
            return self.__system
        else:
            raise PipelineViewError("system not specified in deployment data")

    @system.setter
    def system(self: Y, v: str) -> None:
        """Set system name property."""
        self.__system = v

    def __str__(self: Y) -> str:
        """Convert object to str for logging purposes."""
        return str(
            {
                "commit_sha": self.commit_sha,
                "pipeline_id": self.pipeline_id,
                "release_id": self.release_id,
                "release_state": self.release_state,
                "application_name": self.__application_name,
                "azure_subscription_name": self.__azure_subscription_name,
                "azure_tenant_name": self.__azure_tenant_name,
                "client": self.__client,
                "frame_name": self.__frame_name,
                "system": self.__system,
            }
        )


X = typing.TypeVar("X", bound="DeployDataView")


class DeployDataView:
    """Data critical to a resource deployment."""

    azure_credentials: AzureCredentials
    deployment_tuple: str
    location_primary: str
    release_state: str
    static_secrets: dict

    frame_folder: typing.Optional[pathlib.Path] = None

    __location_secondary: typing.Optional[str] = None

    iteration_context: IterationContext
    to: StructuredTo
    __puff_map: typing.Optional[PuffMap] = None

    def __init__(
        self: X,
        azure_credentials: AzureCredentials,
        deployment_tuple: str,
        location_primary: str,
        release_state: str,
        static_secrets: dict,
        location_secondary: typing.Optional[str] = None,
    ) -> None:
        """Construct ``DeployDataView`` object."""
        self.azure_credentials = azure_credentials
        self.deployment_tuple = deployment_tuple
        self.location_primary = location_primary
        self.release_state = release_state
        self.static_secrets = static_secrets
        self.__location_secondary = location_secondary

        self.iteration_context = IterationContext()
        self.to = StructuredTo()

    @property
    def location_secondary(self: X) -> str:
        """Get location secondary property."""
        if self.__location_secondary:
            return self.__location_secondary
        else:
            raise PipelineViewError(
                "location secondary not specified in " "deployment data"
            )

    @location_secondary.setter
    def location_secondary(self: X, v: str) -> None:
        """Set location secondary property."""
        self.__location_secondary = v

    @property
    def puff_map(self: X) -> PuffMap:
        """Get puff map property."""
        if self.__puff_map:
            return self.__puff_map
        else:
            raise PipelineViewError("puff map not specified in deployment data")

    @puff_map.setter
    def puff_map(self: X, value: PuffMap) -> None:
        """Set puff map property."""
        self.__puff_map = value

    def __str__(self: X) -> str:
        """Convert object to str for logging purposes."""
        return str(
            {
                "azure_credentials": self.azure_credentials,
                "deployment_tuple": self.deployment_tuple,
                "location_primary": self.location_primary,
                "release_state": self.release_state,
                "location_secondary": self.__location_secondary,
                "iteration_context": self.iteration_context,
            }
        )


W = typing.TypeVar("W", bound="FlattenedDeployment")


@dataclasses.dataclass
class FlattenedDeployment:
    """Flattened deployment data for deployment concurrency."""

    context: DeploymentContext
    data: DeployDataView

    def copy_add_frame(self: W, frame_name: str) -> W:
        """
        Deep copy this object and add frame name data.

        Args:
            frame_name: Frame name data to add.

        Returns:
            Copied and updated object.
        """
        x = copy.deepcopy(self)
        x.context.frame_name = frame_name
        x.data.iteration_context.append(frame_name)
        return x

    def copy_add_application(self: W, application_name: str) -> W:
        """
        Deep copy this object and add application name data.

        Args:
            application_name: Application name data to add.

        Returns:
            Copied and updated object.
        """
        x = copy.deepcopy(self)
        x.context.application_name = application_name
        x.data.iteration_context.append(application_name)
        return x


V = typing.TypeVar("V", bound="SubscriptionView")
U = typing.TypeVar("U", bound="DeploymentView")
T = typing.TypeVar("T", bound="ReleaseView")


class SubscriptionView:
    """View of pipeline configuration conditioned by a subscription name."""

    deployment_view: "DeploymentView"
    subscription_name: str

    def __init__(self: V, deployment_view: U, subscription_name: str) -> None:
        """Construct ``SubscriptionView`` object."""
        self.deployment_view = deployment_view
        self.subscription_name = subscription_name

        self._validate_subscription()

    @property
    def deploy_data(self: V) -> typing.List[DeployDataView]:
        """Provide deployment data for each location in this subscription."""
        base_data = (
            self.deployment_view.release_view.configuration.subscriptions[
                self.subscription_name
            ]
        )
        result: typing.List[DeployDataView] = list()
        this_deployment = (
            self.deployment_view.release_view.configuration.deployments[
                str(self.deployment_view.deployment_tuple)
            ]
        )
        this_service_principals = (
            self.deployment_view.release_view.configuration.service_principals  # noqa E501
        )
        if not this_service_principals:
            raise PipelineViewError("Missing service principal credentials")

        this_tenants = self.deployment_view.release_view.configuration.tenants
        for this_locations in this_deployment.subscriptions[
            self.subscription_name
        ].locations:
            this_credentials = AzureCredentials(
                userid=this_service_principals[self.subscription_name].id,
                secret=this_service_principals[self.subscription_name].secret,
                name=this_service_principals[self.subscription_name].name,
                subscription=base_data.azure_id,
                tenant=this_tenants[base_data.tenant].azure_id,
            )
            static_secrets = dict()
            if self.deployment_view.release_view.configuration.static_secrets:
                static_secrets = self.deployment_view.release_view.configuration.static_secrets[  # noqa E501
                    self.subscription_name
                ]

            this_data: typing.Dict[str, typing.Any] = {
                "azure_credentials": this_credentials,
                "deployment_tuple": str(self.deployment_view.deployment_tuple),
                "location_primary": this_locations.primary,
                "location_secondary": None,
                "release_state": self.deployment_view.release_view.deployment_context.release_state,  # noqa E501
                "static_secrets": static_secrets,
            }
            if this_locations.secondary:
                this_data["location_secondary"] = this_locations.secondary

            result.append(DeployDataView(**this_data))

        return result

    def _validate_subscription(self: V) -> None:
        if (
            self.subscription_name
            not in self.deployment_view.release_view.configuration.subscriptions
        ):
            raise PipelineViewError(
                "Bad subscription, {0}".format(str(self.subscription_name))
            )


class DeploymentView:
    """View of pipeline configuration conditioned by deployment state."""

    release_view: "ReleaseView"
    deployment_tuple: DeploymentTuple

    def __init__(
        self: U, release_view: T, deployment_state: DeploymentTuple
    ) -> None:
        """Construct ``DeploymentView`` object."""
        self.release_view = release_view
        self.deployment_tuple = deployment_state

        self._validate_deployment_tuple()

    @property
    def subscriptions(self: U) -> typing.List[SubscriptionView]:
        """Provide the subscriptions in this deployment."""
        result: typing.List[SubscriptionView] = list()
        if (
            str(self.deployment_tuple)
            in self.release_view.configuration.deployments
        ):
            for (
                this_subscription
            ) in self.release_view.configuration.deployments[
                str(self.deployment_tuple)
            ].subscriptions.keys():
                result.append(SubscriptionView(self, this_subscription))

        return result

    def _validate_deployment_tuple(self: U) -> None:
        if (
            self.deployment_tuple.release_state
            not in self.release_view.configuration.release_states
        ):
            raise PipelineViewError(
                "Bad release state, {0}".format(str(self.deployment_tuple))
            )

        if (
            self.deployment_tuple.client
            not in self.release_view.configuration.clients.keys()
        ):
            raise PipelineViewError(
                "Bad client, {0}".format(str(self.deployment_tuple))
            )

        if (
            self.deployment_tuple.system
            not in self.release_view.configuration.systems
        ):
            raise PipelineViewError(
                "Bad system, {0}".format(str(self.deployment_tuple))
            )


class ReleaseView:
    """View of pipeline configuration conditioned by release state."""

    configuration: PipelineConfiguration
    deployment_context: DeploymentContext

    def __init__(
        self: T,
        configuration: PipelineConfiguration,
        deployment_context: DeploymentContext,
    ) -> None:
        """Construct ``ReleaseView`` object."""
        self.configuration = configuration
        self.deployment_context = deployment_context

        self._validate_release_state()

    @property
    def deployments(self: T) -> typing.List[DeploymentView]:
        """Provide the deployments in this release."""
        result: typing.List[DeploymentView] = list()
        for this_client in self.configuration.clients.keys():
            for this_system in self.configuration.systems:
                this_state = DeploymentTuple(
                    client=this_client,
                    release_state=self.deployment_context.release_state,
                    system=this_system,
                )
                if str(this_state) in self.configuration.deployments.keys():
                    result.append(DeploymentView(self, this_state))

        return result

    def flatten(
        self: T,
        to: StructuredTo,
    ) -> typing.List[FlattenedDeployment]:
        """Flatten the nested hierarchy of views into a simple list."""
        result: typing.List[FlattenedDeployment] = list()
        for this_deployment in self.deployments:
            for this_subscription in this_deployment.subscriptions:
                for this_deploy_data in this_subscription.deploy_data:
                    updated_context = copy.deepcopy(self.deployment_context)
                    updated_context.azure_subscription_name = (
                        this_subscription.subscription_name
                    )
                    updated_context.azure_tenant_name = (
                        self.configuration.subscriptions[
                            this_subscription.subscription_name
                        ].tenant
                    )
                    updated_context.client = (
                        this_deployment.deployment_tuple.client
                    )
                    updated_context.system = (
                        this_deployment.deployment_tuple.system
                    )

                    updated_data = copy.deepcopy(this_deploy_data)
                    updated_data.to = to

                    this_value = FlattenedDeployment(
                        context=updated_context,
                        data=updated_data,
                    )
                    result.append(this_value)
        return result

    def _validate_release_state(self: T) -> None:
        if (
            self.deployment_context.release_state
            not in self.configuration.release_states
        ):
            message = "Bad release state, {0}, {1}".format(
                self.deployment_context.release_state,
                str(self.configuration.release_states),
            )
            raise PipelineViewError(message)



# Copyright (c) 2021 Food-X Technologies
#
# This file is part of foodx_devops_tools.
#
# You should have received a copy of the MIT License along with
# foodx_devops_tools. If not, see <https://opensource.org/licenses/MIT>.

import pytest

from foodx_devops_tools.release_flow._cd_release_state import (
    ReleaseState,
    ReleaseStateError,
    identify_release_state,
)


class TestIdentifyReleaseState:
    def test_clean_feature(self):
        expected_state = ReleaseState.ftr

        result = identify_release_state("refs/heads/feature/some/path")
        assert result == expected_state

        # Assume PR is only valid for feature/*->main
        result = identify_release_state("refs/pull/1/merge")
        assert result == expected_state

    def test_clean_development(self):
        expected_state = ReleaseState.dev

        result = identify_release_state("refs/heads/main")
        assert result == expected_state

    def test_clean_qa(self):
        expected_state = ReleaseState.qa

        result = identify_release_state("refs/tags/3.14.159-alpha.13")
        assert result == expected_state

    def test_clean_staging(self):
        expected_state = ReleaseState.stg

        result = identify_release_state("refs/tags/3.14.159-beta.13")
        assert result == expected_state

    def test_clean_production(self):
        expected_state = ReleaseState.prd

        result = identify_release_state("refs/tags/3.14.159")
        assert result == expected_state

    def test_dirty_qa(self):
        """Only strictly semantic version alpha tags allowed."""
        with pytest.raises(
            ReleaseStateError,
            match="^Unable to match git reference to any release state",
        ):
            identify_release_state("refs/tags/3.14.159-alpha")

        with pytest.raises(
            ReleaseStateError,
            match="^Unable to match git reference to any release state",
        ):
            identify_release_state("refs/tags/3.14.159-alpha13")

    def test_dirty_staging(self):
        """Only strictly semantic version beta tags allowed."""
        with pytest.raises(
            ReleaseStateError,
            match="^Unable to match git reference to any release state",
        ):
            identify_release_state("refs/tags/3.14.159-beta")

        with pytest.raises(
            ReleaseStateError,
            match="^Unable to match git reference to any release state",
        ):
            identify_release_state("refs/tags/3.14.159-beta13")

    def test_dirty_production(self):
        """Only strictly semantic releases are allowed."""
        with pytest.raises(
            ReleaseStateError,
            match="^Unable to match git reference to any release state",
        ):
            identify_release_state("refs/tags/3.14.159.54")



#  Copyright (c) 2021 Food-X Technologies
#
#  This file is part of foodx_devops_tools.
#
#  You should have received a copy of the MIT License along with
#  foodx_devops_tools. If not, see <https://opensource.org/licenses/MIT>.

"""Azure ARM template parameter generation."""

import copy
import json
import logging
import pathlib
import typing

import aiofiles  # type: ignore
import aiofiles.os  # type: ignore
import click
import pydantic
import ruamel.yaml
from deepmerge import always_merger  # type: ignore

from ._exceptions import ArmTemplateError
from ._header import ARMTEMPLATE_PARAMETERS_HEADER
from ._puff_parameters import PuffParameterModel

log = logging.getLogger(__name__)

YamlData = typing.Dict[str, typing.Any]

BASE_PARAMETER_EXCLUDE_LABELS = ["default", "environments", "name", "services"]


async def load_yaml(path: pathlib.Path) -> YamlData:
    """
    Asynchronously load yaml file data.

    Args:
        path: Path to YAML file to load.

    Returns:
        Loaded YAML data.
    Raises:
        FileNotFoundError: If file cannot be loaded.
        IsADirectoryError: If path is a directory instead of a file.
    """
    try:
        yaml = ruamel.yaml.YAML(typ="safe")
        async with aiofiles.open(str(path), mode="r") as f:
            content = await f.read()
        yaml_data = yaml.load(content)
        log.debug("{0}, {1}".format(path, str(yaml_data)))

        if not yaml_data:
            log.warning("Empty YAML file, {0}".format(str(path)))

        return yaml_data
    except FileNotFoundError:
        log.error("File not found, {0}".format(str(path)))
        raise
    except IsADirectoryError:
        log.error("Path is a directory, {0}".format(str(path)))
        raise
    except Exception:
        log.error("Unexpected error with path, {0}".format(str(path)))
        raise


def _remove_keys(
    parameters: typing.Dict[str, typing.Any], exclude_labels: typing.List[str]
) -> dict:
    """
    Remove keys from a dictionary without changing the original.

    Attempts to remove keys that don't exist in the dictinary are silently
    ignored.

    Args:
        parameters: Dictionary to be adjusted.

    Returns:
        Modified dictionary.
    """
    this_copy = parameters.copy()
    for k in exclude_labels:
        try:
            del this_copy[k]
        except KeyError:
            # silently ignore missing keys by suppressing this exception
            pass

    return this_copy


async def _save_parameter_file(
    target_path: pathlib.Path,
    parameter_data: typing.Optional[dict],
    is_pretty: bool,
) -> None:
    """
    Create ARM template parameter file.

    Args:
        target_path: File path to be created or deleted.
        parameter_data: Parameters to be applied to data.
        is_pretty: Create nicely formatted JSON for humans.
    """
    # WARNING: The header template dict contains a dict in a field and
    # dict.copy() is shallow so care needs to be taken to ensure all
    # concurrent copies are unique due to Python reference-copy default.
    generated_parameters = copy.deepcopy(ARMTEMPLATE_PARAMETERS_HEADER)
    if parameter_data:
        parameters = dict()
        for item_key, value in parameter_data.items():
            if isinstance(value, dict):
                # assume the dict is valid ARM template parameter structure.
                # the dict SHOULD be an ARM template parameter reference:
                # https://docs.microsoft.com/en-us/azure/azure-resource-manager/templates/parameter-files#parameter-file
                parameters[item_key] = value
            elif isinstance(value, (bool, float, int, str, list)):
                parameters[item_key] = {
                    "value": value,
                }
            else:
                # assume the value is convertible to str
                parameters[item_key] = {
                    "value": str(value),
                }
        generated_parameters["parameters"] = parameters
    dump_arguments: dict = dict()
    if is_pretty:
        dump_arguments = {"sort_keys": True, "indent": 2}
    async with aiofiles.open(target_path, mode="w") as f:
        log.info("saving parameter file, {0}".format(target_path))
        await f.write(json.dumps(generated_parameters, **dump_arguments))


async def _delete_parameter_file(
    target_path: pathlib.Path,
) -> None:
    """
    Delete generated ARM template parameter file.

    Args:
        target_path: File path to be created or deleted.
    """
    if target_path.exists():
        log.warning("deleting parameter file, {0}".format(target_path))
        await aiofiles.os.remove(target_path)


async def _do_file_actions(
    is_delete_action: bool,
    target_path: pathlib.Path,
    parameter_data: dict,
    is_pretty: bool,
) -> None:
    """
    Create or delete generated ARM template parameter files.

    Args:
        is_delete_action: Signal deleting or creating parameter files.
        target_path: Directory to store or delete parameter files.
        parameter_data: Data for each ARM template parameter file.
        is_pretty: Create nicely formatted JSON for humans.
    """
    for key, values in parameter_data.items():
        this_path = target_path / ".".join([key, "json"])
        if is_delete_action:
            await _delete_parameter_file(this_path)
        else:
            await _save_parameter_file(this_path, values, is_pretty)


def _linearize_name(base_data: dict, filename: str) -> dict:
    cleaned = _remove_keys(base_data, ["name"])
    this_name = base_data.get("name", filename)
    if not this_name:
        raise RuntimeError("Invalid empty name")

    linearized = {
        this_name: cleaned,
    }
    return linearized


def _merge_default_name(base_data: dict) -> dict:
    this_data = copy.deepcopy(base_data)
    for key, value in this_data.items():
        if "default" in value:
            new_parameters = value["default"]
            cleaned = _remove_keys(value, ["default"])

            updated = always_merger.merge(cleaned, new_parameters)

            this_data[key] = updated

    return this_data


def _iterate_iterable(
    cleaned: dict, base_key: str, this_data: dict, singular: str
) -> dict:
    linearized = dict()
    for iterable_key, iterable_data in this_data.items():
        if not iterable_key:
            raise RuntimeError("Invalid empty {0} name".format(singular))

        if iterable_data:
            this_value = always_merger.merge(
                copy.deepcopy(cleaned), iterable_data
            )
        else:
            this_value = copy.deepcopy(cleaned)

        this_value[singular] = iterable_key
        linearized["{0}.{1}".format(base_key, iterable_key)] = this_value

    return linearized


def _linearize_iterable(base_data: dict, plural: str, singular: str) -> dict:
    linearized = dict()
    for base_key, base_value in base_data.items():
        this_data = (
            copy.deepcopy(base_value[plural])
            if plural in base_value
            else dict()
        )
        cleaned = _remove_keys(copy.deepcopy(base_value), [plural])
        if this_data:
            iterated = _iterate_iterable(cleaned, base_key, this_data, singular)
            linearized.update(iterated)
        else:
            linearized[base_key] = cleaned

    return linearized if linearized else base_data.copy()


def _linearize_services(base_data: dict) -> dict:
    plural = "services"
    singular = "service"
    return _linearize_iterable(base_data, plural, singular)


def _linearize_environments(base_data: dict) -> dict:
    plural = "environments"
    singular = "environment"
    return _linearize_iterable(base_data, plural, singular)


def _merge_list_of_dict(this_data: typing.List[dict]) -> dict:
    result: dict = dict()
    for item in this_data:
        if isinstance(item, dict):
            for k, v in item.items():
                if k in result:
                    result[k].update(v if v is not None else dict())
                else:
                    result[k] = v if v is not None else dict()
        else:
            log.error("ignoring non-dict list entry, {0}".format(str(item)))
    return result


def _linearize_regions(base_data: dict) -> dict:
    plural = "regions"
    singular = "region"
    linearized = dict()
    for base_key, base_value in base_data.items():
        this_data = (
            copy.deepcopy(base_value[plural])
            if plural in base_value
            else list()
        )
        cleaned = _remove_keys(base_value, [plural])
        # special handling of the region list unique to regions
        if this_data:
            merged_list = _merge_list_of_dict(this_data)
            iterated = _iterate_iterable(
                cleaned, base_key, merged_list, singular
            )
            linearized.update(iterated)
        else:
            linearized[base_key] = cleaned

    return linearized if linearized else copy.deepcopy(base_data)


def _linearize_parameters(yaml_data: dict, file_name: str) -> dict:
    level0_merge = _linearize_name(yaml_data, file_name)

    # a hack to enable a variable called "name" to be specified.
    default_merged = _merge_default_name(level0_merge)

    level1_merge = _linearize_services(default_merged)
    level2_merge = _linearize_environments(level1_merge)
    level3_merge = _linearize_regions(level2_merge)

    return level3_merge


async def do_arm_template_parameter_action(
    puff_file_path: pathlib.Path,
    is_delete_action: bool,
    is_pretty: bool,
) -> None:
    """
    Generate ARM template parameter files from puff YAML file.

    The generated files are always output to the same directory as the source
    puff file.

    Args:
        puff_file_path: Path to source YAML parameter file.
        is_delete_action: True if files should be deleted instead of created.
        is_pretty: Create nicely formatted JSON for humans.
    """
    target_path = puff_file_path.parent

    click.echo("loading, {0}".format(puff_file_path))
    yaml_data = await load_yaml(puff_file_path)
    try:
        # for now, just use the pydantic model to validate the YAML data.
        PuffParameterModel.parse_obj(yaml_data)

        file_name = pathlib.Path(puff_file_path.stem).name
        merged_data = _linearize_parameters(yaml_data, file_name)
        await _do_file_actions(
            is_delete_action, target_path, merged_data, is_pretty
        )
    except pydantic.ValidationError as e:
        raise ArmTemplateError(
            "Puff parameter validation failed, "
            "{0}. {1}".format(puff_file_path, str(e))
        ) from e



#  Copyright (c) 2021 Food-X Technologies
#
#  This file is part of foodx_devops_tools.
#
#  You should have received a copy of the MIT License along with
#  foodx_devops_tools. If not, see <https://opensource.org/licenses/MIT>.

import copy
import logging
import pathlib

import pytest

from foodx_devops_tools._to import StructuredTo
from foodx_devops_tools.deploy_me._deployment import (
    ApplicationDeploymentDefinition,
    DeploymentState,
    DeploymentStatus,
    FlattenedDeployment,
    _deploy_step,
)
from foodx_devops_tools.pipeline_config.frames import DeploymentMode
from foodx_devops_tools.pipeline_config.views import (
    AzureCredentials,
    DeployDataView,
)
from tests.ci.support.pipeline_config import MOCK_CONTEXT


@pytest.fixture()
def mock_rg_deploy(mock_async_method):
    this_mock = mock_async_method(
        "foodx_devops_tools.deploy_me._deployment" ".deploy_resource_group"
    )
    return this_mock


@pytest.fixture()
def mock_login(mock_async_method):
    mock_async_method(
        "foodx_devops_tools.deploy_me._deployment" ".login_service_principal"
    )


@pytest.fixture()
def mock_deploystep_context():
    mock_context = copy.deepcopy(MOCK_CONTEXT)
    mock_context.application_name = "app-name"
    mock_context.azure_tenant_name = "t1"
    mock_context.azure_subscription_name = "sb1"
    mock_context.client = "c1"
    mock_context.system = "sys1"
    mock_context.frame_name = "f1"

    data_view = DeployDataView(
        AzureCredentials(
            userid="abc",
            secret="verysecret",
            subscription="abc123",
            name="n",
            tenant="123abc",
        ),
        "a-b-c",
        "uswest2",
        "r1",
        dict(),
    )
    data_view.frame_folder = pathlib.Path("frame/folder")

    arguments = {
        "this_step": ApplicationDeploymentDefinition(
            mode=DeploymentMode.incremental,
            name="this_step",
            arm_file=pathlib.Path("arm.file"),
            puff_file=None,
            resource_group=None,
        ),
        "deployment_data": FlattenedDeployment(
            context=mock_context, data=data_view
        ),
        "puff_parameter_data": {
            "this_step": pathlib.Path("puff.path"),
        },
        "this_context": "some.context",
        "enable_validation": False,
    }
    return arguments


@pytest.mark.asyncio
async def test_clean(mock_deploystep_context, mock_login, mock_rg_deploy):
    await _deploy_step(**mock_deploystep_context)

    mock_rg_deploy.assert_called_once()


@pytest.mark.asyncio
async def test_no_override_parameters(
    mock_deploystep_context, mock_login, mock_rg_deploy, mocker
):
    this_context = copy.deepcopy(mock_deploystep_context)
    this_context["this_step"].static_secrets = False

    await _deploy_step(**this_context)

    mock_rg_deploy.assert_called_once_with(
        mocker.ANY,
        mocker.ANY,
        mocker.ANY,
        mocker.ANY,
        mocker.ANY,
        mocker.ANY,
        override_parameters=dict(),
        validate=mocker.ANY,
    )


@pytest.mark.asyncio
async def test_secrets_enabled(
    mock_deploystep_context, mock_login, mock_rg_deploy, mocker
):
    this_context = copy.deepcopy(mock_deploystep_context)
    this_context["this_step"].static_secrets = True

    this_context["deployment_data"].data.static_secrets = {
        "one": 1,
        "two": "2",
    }
    expected_secrets = {
        "staticSecrets": [
            {
                "enabled": True,
                "key": "one",
                "value": 1,
            },
            {
                "enabled": True,
                "key": "two",
                "value": "2",
            },
        ],
    }

    await _deploy_step(**this_context)

    mock_rg_deploy.assert_called_once_with(
        mocker.ANY,
        mocker.ANY,
        mocker.ANY,
        mocker.ANY,
        mocker.ANY,
        mocker.ANY,
        override_parameters=expected_secrets,
        validate=mocker.ANY,
    )


@pytest.mark.asyncio
async def test_step_skip(
    caplog, mock_deploystep_context, mock_login, mock_rg_deploy
):
    with caplog.at_level(logging.DEBUG):
        mock_deploystep_context["deployment_data"].data.to = StructuredTo(
            frame="f1", application="a1", step="other_step"
        )

        await _deploy_step(**mock_deploystep_context)

        mock_rg_deploy.assert_not_called()

        assert (
            "application step skipped using deployment specifier" in caplog.text
        )



#  Copyright (c) 2021 Food-X Technologies
#
#  This file is part of foodx_devops_tools.
#
#  You should have received a copy of the MIT License along with
#  foodx_devops_tools. If not, see <https://opensource.org/licenses/MIT>.

import contextlib
import os
import pathlib
import tempfile
import typing

import ruamel.yaml

from foodx_devops_tools._to import StructuredTo
from foodx_devops_tools.pipeline_config import (
    DeploymentContext,
    PipelineConfigurationPaths,
)
from foodx_devops_tools.utilities.ansible import _encrypt_vault

MOCK_CONTEXT = DeploymentContext(
    commit_sha="abc123",
    pipeline_id="12345",
    release_id="0.0.0-dev.3",
    release_state="r1",
)


MOCK_TO = StructuredTo()

CLEAN_SPLIT = {
    "client": {
        "clients",
        "deployments",
        "frames",
        "puff_map",
        "service_principals",
        "static_secrets",
    },
    "system": {
        "release_states",
        "subscriptions",
        "systems",
        "tenants",
    },
}
NOT_SPLIT = {
    "client": {},
    "system": {
        "release_states",
        "subscriptions",
        "systems",
        "tenants",
        "clients",
        "deployments",
        "frames",
        "puff_map",
        "service_principals",
        "static_secrets",
    },
}

MOCK_SECRET = "verysecret"

MOCK_RESULTS = {
    "clients": {
        "c1": {
            "release_states": [
                "r1",
                "r2",
            ],
            "system": "sys1",
        },
        "c2": {
            "release_states": [
                "r2",
            ],
            "system": "sys2",
        },
    },
    "deployments": {
        "sys1-c1-r1": {
            "subscriptions": {
                "sub1": {
                    "locations": [{"primary": "l1"}, {"primary": "l2"}],
                },
            },
        },
    },
    "frames": {
        "frames": {
            "f1": {
                "applications": {
                    "a1": [
                        {
                            "resource_group": "a1_group",
                            "name": "a1l1",
                            "mode": "Incremental",
                        },
                    ],
                },
                "folder": "some/path",
            },
        },
    },
    "puff_map": {
        "frames": {
            "f1": {
                "applications": {
                    "a1": {
                        "arm_parameters_files": {
                            "r1": {
                                "sub1": {
                                    "a1l1": "some/puff_map/path",
                                },
                            },
                        },
                    },
                },
            },
        },
    },
    "release_states": ["r1", "r2"],
    "service_principals": {
        "sub1": {"id": "12345", "secret": "verysecret", "name": "sp_name"},
    },
    "static_secrets": {
        "sub1": {"k1": "k1v"},
    },
    "subscriptions": {
        "sub1": {
            "ado_service_connection": "some-name",
            "azure_id": "abc123",
            "tenant": "t1",
        },
    },
    "systems": ["sys1", "sys2"],
    "tenants": {"t1": {"azure_id": "123abc"}},
}

MOCK_PATHS = PipelineConfigurationPaths.from_dict(
    {
        "clients": pathlib.Path("client/path"),
        "release_states": pathlib.Path("release_state/path"),
        "deployments": pathlib.Path("deployment/path"),
        "frames": pathlib.Path("frame/path"),
        "puff_map": pathlib.Path("puff_map/path"),
        "service_principals": pathlib.Path("service_principal/path"),
        "static_secrets": {pathlib.Path("static_secrets/path")},
        "subscriptions": pathlib.Path("subscription/path"),
        "systems": pathlib.Path("system/path"),
        "tenants": pathlib.Path("tenant/path"),
    }
)


def _create_encrypted_file(
    file_path: pathlib.Path,
    data: dict,
    decrypt_token: str,
    yaml_label: typing.Optional[str] = None,
):
    yaml = ruamel.yaml.YAML(typ="safe")
    this_file = file_path.stem
    if not yaml_label:
        yaml_label = file_path.stem
    with file_path.open(mode="w") as f:
        yaml.dump({yaml_label: data}, f)

    encrypted_file_path = file_path.parent / "{0}.vault".format(this_file)
    password_file_path = pathlib.Path(str(encrypted_file_path) + ".password")
    with password_file_path.open(mode="w") as f:
        f.write(decrypt_token)
    _encrypt_vault(encrypted_file_path, password_file_path, file_path)

    os.remove(password_file_path)
    os.remove(file_path)


@contextlib.contextmanager
def split_directories(split: dict, decrypt_token: str = MOCK_SECRET):
    """Generate configuration files split across two directories."""
    yaml = ruamel.yaml.YAML(typ="safe")
    with tempfile.TemporaryDirectory() as dir1:
        pd1 = pathlib.Path(dir1)
        with tempfile.TemporaryDirectory() as dir2:
            pd2 = pathlib.Path(dir2)
            paths = {
                "client": pd1,
                "system": pd2,
            }
            for key, files in split.items():
                for this_file in files:
                    if this_file not in [
                        "service_principals",
                        "static_secrets",
                    ]:
                        file_path = paths[key] / "{0}.yml".format(this_file)
                        with file_path.open(mode="w") as f:
                            yaml.dump({this_file: MOCK_RESULTS[this_file]}, f)
                    elif this_file == "service_principals":
                        # service_principals requires special handling due to
                        # encryption
                        file_path = paths[key] / "{0}.yml".format(this_file)
                        _create_encrypted_file(
                            file_path, MOCK_RESULTS[this_file], decrypt_token
                        )
                    elif this_file == "static_secrets":
                        # static_secrets require special handling due to
                        # encryption and a directory of files.
                        this_dir = paths[key] / this_file
                        os.makedirs(this_dir, exist_ok=True)
                        for name, data in MOCK_RESULTS[this_file].items():
                            file_path = this_dir / "{0}.yml".format(name)
                            _create_encrypted_file(
                                file_path,
                                MOCK_RESULTS[this_file],
                                decrypt_token,
                                yaml_label=this_file,
                            )
                    else:
                        raise RuntimeError(
                            "unknown file type, {0}".format(this_file)
                        )

            yield pd1, pd2



#  Copyright (c) 2021 Food-X Technologies
#
#  This file is part of foodx_devops_tools.
#
#  You should have received a copy of the MIT License along with
#  foodx_devops_tools. If not, see <https://opensource.org/licenses/MIT>.

"""Export exception classes for public consumption."""

from ._exceptions import (  # noqa: F401
    ClientsDefinitionError,
    ConfigurationPathsError,
    DeploymentsDefinitionError,
    FolderPathError,
    FrameDefinitionsError,
    PipelineConfigurationError,
    PipelineViewError,
    PuffMapDefinitionsError,
    ReleaseStatesDefinitionError,
    SubscriptionsDefinitionError,
    SystemsDefinitionError,
    TenantsDefinitionError,
)



#  Copyright (c) 2021 Food-X Technologies
#
#  This file is part of foodx_devops_tools.
#
#  You should have received a copy of the MIT License along with
#  foodx_devops_tools. If not, see <https://opensource.org/licenses/MIT>.

"""Ansible related utilities."""

import contextlib
import logging
import os
import pathlib
import typing

from ._exceptions import AnsibleVaultError
from .command import run_command

log = logging.getLogger(__name__)


def _encrypt_vault(
    encrypted_file_path: pathlib.Path,
    password_file_path: pathlib.Path,
    unencrypted_file_path: pathlib.Path,
) -> None:
    this_command = [
        "ansible-vault",
        "encrypt",
        "--output",
        str(encrypted_file_path),
        "--vault-password-file",
        str(password_file_path),
        str(unencrypted_file_path),
    ]
    result = run_command(this_command)
    if result.returncode != 0:
        message = "Error encrypting Ansible vault file, {0}".format(
            encrypted_file_path
        )
        log.debug("stdout, {0}".format(result.stdout))
        log.debug("stderr, {0}".format(result.stderr))
        raise AnsibleVaultError(message)


def _decrypt_vault(
    encrypted_file_path: pathlib.Path,
    decrypted_path: pathlib.Path,
    password_file_path: pathlib.Path,
) -> None:
    this_command = [
        "ansible-vault",
        "decrypt",
        "--output",
        str(decrypted_path),
        "--vault-password-file",
        str(password_file_path),
        str(encrypted_file_path),
    ]
    result = run_command(this_command)
    if result.returncode != 0:
        message = "Error decrypting Ansible vault file, {0}".format(
            encrypted_file_path
        )
        log.debug("stdout, {0}".format(result.stdout))
        log.debug("stderr, {0}".format(result.stderr))
        raise AnsibleVaultError(message)


@contextlib.contextmanager
def managed_file_decrypt(
    encrypted_file_path: pathlib.Path, decrypt_token: str
) -> typing.Generator[typing.TextIO, None, None]:
    """
    Manage the decryption of an Ansible vault within a context.

    The encrypted Ansible vault is decrypted to an unencrypted file for the
    duration of the context. On context exit the unencrypted file is deleted.

    Args:
        encrypted_file_path: Path to encrypted file.
        decrypt_token: Token for decrypting Ansible vault.

    Yields:
        File stream object of decrypted file.
    """
    decrypted_path = pathlib.Path(str(encrypted_file_path) + ".yml")
    password_file_path = pathlib.Path(str(encrypted_file_path) + ".password")
    try:
        with password_file_path.open(mode="w") as f:
            f.write(decrypt_token)
        _decrypt_vault(encrypted_file_path, decrypted_path, password_file_path)
        os.remove(password_file_path)

        with decrypted_path.open(mode="r") as f:
            yield f
    except AnsibleVaultError:
        raise
    except Exception as e:
        message = "Failed to decrypt Ansible vault file, {0}".format(str(e))
        log.error(message)
        raise AnsibleVaultError(message) from e
    finally:
        if decrypted_path.exists():
            os.remove(decrypted_path)
        if password_file_path.exists():
            os.remove(password_file_path)



#  Copyright (c) 2021 Food-X Technologies
#
#  This file is part of foodx_devops_tools.
#
#  You should have received a copy of the MIT License along with
#  foodx_devops_tools. If not, see <https://opensource.org/licenses/MIT>.

from foodx_devops_tools.pipeline_config.exceptions import (
    PipelineConfigurationError,
)
from foodx_devops_tools.validate_configuration import ExitState, _main
from tests.ci.support.click_runner import (  # noqa: F401
    click_runner,
    isolated_filesystem,
)
from tests.ci.support.pipeline_config import (
    CLEAN_SPLIT,
    MOCK_SECRET,
    NOT_SPLIT,
    split_directories,
)


class TestMain:
    def test_default(self, click_runner):
        with split_directories(NOT_SPLIT.copy()) as (
            client_config,
            system_config,
        ):
            result = click_runner.invoke(
                _main,
                [str(client_config), str(system_config), "-"],
                input=MOCK_SECRET,
            )

            assert result.exit_code == 0

    def test_missing_input_ignored(self, click_runner):
        """--disable-vaults option doesn't read from stdin."""
        with split_directories(NOT_SPLIT.copy()) as (
            client_config,
            system_config,
        ):
            result = click_runner.invoke(
                _main,
                [
                    str(client_config),
                    str(system_config),
                    "-",
                    "--disable-vaults",
                ],
            )

            assert result.exit_code == 0

    def test_split_files(self, click_runner):
        with split_directories(CLEAN_SPLIT.copy()) as (
            client_config,
            system_config,
        ):
            result = click_runner.invoke(
                _main,
                [
                    str(client_config),
                    str(system_config),
                    "-",
                ],
                input=MOCK_SECRET,
            )

            assert result.exit_code == 0

    def test_configuration_error_exits_dirty(self, click_runner, mocker):
        expected_dir = "some/path"

        mocker.patch(
            "foodx_devops_tools.validate_configuration"
            ".PipelineConfiguration.from_files",
            side_effect=PipelineConfigurationError("some error"),
        )

        with split_directories(CLEAN_SPLIT.copy()) as (
            client_config,
            system_config,
        ):
            result = click_runner.invoke(
                _main,
                [
                    str(client_config),
                    str(system_config),
                    "-",
                ],
                input=MOCK_SECRET,
            )

            assert result.exit_code == ExitState.MISSING_GITREF.value

    def test_unknown_error_exits_dirty(self, click_runner, mocker):
        expected_dir = "some/path"

        mocker.patch(
            "foodx_devops_tools.validate_configuration"
            ".PipelineConfiguration.from_files",
            side_effect=RuntimeError("some error"),
        )

        with split_directories(CLEAN_SPLIT.copy()) as (
            client_config,
            system_config,
        ):
            result = click_runner.invoke(
                _main,
                [
                    str(client_config),
                    str(system_config),
                    "-",
                ],
                input=MOCK_SECRET,
            )

            assert result.exit_code == ExitState.UNKNOWN.value



#  Copyright (c) 2021 Food-X Technologies
#
#  This file is part of foodx_devops_tools.
#
#  You should have received a copy of the MIT License along with
#  foodx_devops_tools. If not, see <https://opensource.org/licenses/MIT>.

import logging
import pathlib

from foodx_devops_tools.utilities.ansible import managed_file_decrypt
from tests.ci.support.ansible import encrypted_file


class TestManagedFile:
    def test_clean(self, caplog):
        with caplog.at_level(logging.DEBUG):
            file_text = """
    ---
    service_principals:
      sub1_name:
        id: 12345-id
        name: principal-name1
        secret: verysecret
    """
            decrypt_token = "somesecret"
            with encrypted_file(file_text, decrypt_token) as (
                encrypted_file_path
            ):
                expected_plain_file = pathlib.Path(
                    str(encrypted_file_path) + ".yml"
                )
                expected_password_file = pathlib.Path(
                    str(encrypted_file_path) + ".password"
                )
                # temporary files do not exist
                assert not expected_plain_file.exists()
                assert not expected_password_file.exists()

                with managed_file_decrypt(
                    encrypted_file_path, decrypt_token
                ) as f:
                    # plain text file has been created
                    assert expected_plain_file.is_file()

                    content = f.read()

                    assert content == file_text

                # temporary files have been erased
                assert not expected_plain_file.exists()
                assert not expected_password_file.exists()



#  Copyright (c) 2021 Food-X Technologies
#
#  This file is part of foodx_devops_tools.
#
#  You should have received a copy of the MIT License along with
#  foodx_devops_tools. If not, see <https://opensource.org/licenses/MIT>.

import pytest

from foodx_devops_tools.deploy_me._deployment import PipelineCliOptions


@pytest.fixture()
def mock_completion_event(mock_async_method):
    mock_async_method(
        "foodx_devops_tools.deploy_me._deployment"
        ".DeploymentStatus.wait_for_all_completed"
    )


@pytest.fixture()
def pipeline_parameters():
    def _apply(
        enable_validation=True,
        monitor_sleep_seconds=30,
        wait_timeout_seconds=10,
    ):
        cli_options = PipelineCliOptions(
            enable_validation=enable_validation,
            monitor_sleep_seconds=monitor_sleep_seconds,
            wait_timeout_seconds=wait_timeout_seconds,
        )

        return cli_options

    return _apply



#  Copyright (c) 2021 Food-X Technologies
#
#  This file is part of foodx_devops_tools.
#
#  You should have received a copy of the MIT License along with
#  foodx_devops_tools. If not, see <https://opensource.org/licenses/MIT>.

"""Service principal secrets deployment configuration I/O."""

import logging
import pathlib
import typing

import pydantic

from foodx_devops_tools.utilities.ansible import AnsibleVaultError
from foodx_devops_tools.utilities.io import load_encrypted_data

from ._exceptions import ServicePrincipalsError

log = logging.getLogger(__name__)


class PrincipalConfiguration(pydantic.BaseModel):
    """Define id and key needed for service principal access."""

    id: str
    secret: str
    name: str


ValueType = typing.Dict[str, PrincipalConfiguration]


class ServicePrincipals(pydantic.BaseModel):
    """Define a collection of service principal id's and keys."""

    service_principals: ValueType


def load_service_principals(
    encrypted_file_path: pathlib.Path, decrypt_token: str
) -> ServicePrincipals:
    """Load service principal secrets from a string."""
    try:
        yaml_data = load_encrypted_data(encrypted_file_path, decrypt_token)
        this_object = ServicePrincipals.parse_obj(yaml_data)

        return this_object
    except AnsibleVaultError as e:
        raise ServicePrincipalsError(str(e)) from e
